# Copyright (c) 2008, Aldo Cortesi. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import typing
from collections import defaultdict

from libqtile import configurable, hook
from libqtile.command.base import CommandObject, expose_command
from libqtile.log_utils import logger
from libqtile.utils import has_transparency, is_valid_colors

if typing.TYPE_CHECKING:
    import asyncio
    from typing import Any

    from libqtile.backend.base import Internal, WindowType
    from libqtile.command.base import ItemT
    from libqtile.config import Screen
    from libqtile.core.manager import Qtile
    from libqtile.utils import ColorsType
    from libqtile.widget.base import _Widget


NESW = ("top", "right", "bottom", "left")


class Gap:
    """A gap placed along one of the edges of the screen

    Qtile will avoid covering gaps with windows.

    Parameters
    ==========
    size :
        The "thickness" of the gap, i.e. the height of a horizontal gap, or the
        width of a vertical gap.
    """

    def __init__(self, size: int) -> None:
        # 'size' corresponds to the height of a horizontal gap, or the width
        # of a vertical gap
        self._size = size
        self._initial_size = size
        # '_length' corresponds to the width of a horizontal gap, or the height
        # of a vertical gap
        self._length: int = 0
        self.qtile: Qtile | None = None
        self.screen: Screen | None = None
        self.x: int = 0
        self.y: int = 0
        self.width: int = 0
        self.height: int = 0
        self.horizontal: bool = False

        # Additional reserved around the gap/bar, used when space is dynamically
        # reserved e.g. by third-party bars.
        self.margin: list[int] = [0, 0, 0, 0]  # [N, E, S, W]

    def _configure(self, qtile: Qtile, screen: Screen, reconfigure: bool = False) -> None:
        self.qtile = qtile
        self.screen = screen
        self._size = self._initial_size
        # If both horizontal and vertical gaps are present, screen corners are
        # given to the horizontal ones
        if screen.top is self:
            self.x = screen.x + self.margin[3]
            self.y = screen.y + self.margin[0]
            self._length = screen.width - self.margin[1] - self.margin[3]
            self.width = self._length
            self.height = self._initial_size
            self.horizontal = True
            self._size += self.margin[0] + self.margin[2]
        elif screen.bottom is self:
            self.x = screen.x + self.margin[3]
            self.y = screen.dy + screen.dheight - self.margin[2]
            self._length = screen.width - self.margin[1] - self.margin[3]
            self.width = self._length
            self.height = self._initial_size
            self.horizontal = True
            self._size += self.margin[0] + self.margin[2]
        elif screen.left is self:
            self.x = screen.x + self.margin[3]
            self.y = screen.dy + self.margin[0]
            self._length = screen.dheight - self.margin[0] - self.margin[2]
            self.width = self._initial_size
            self.height = self._length
            self.horizontal = False
            self._size += self.margin[1] + self.margin[3]
        else:  # right
            self.x = screen.dx + screen.dwidth - self.margin[1]
            self.y = screen.dy + self.margin[0]
            self._length = screen.dheight - self.margin[0] - self.margin[2]
            self.width = self._initial_size
            self.height = self._length
            self.horizontal = False
            self._size += self.margin[1] + self.margin[3]

    def draw(self) -> None:
        pass

    def finalize(self) -> None:
        pass

    def geometry(self) -> tuple[int, int, int, int]:
        return (self.x, self.y, self.width, self.height)

    @property
    def size(self) -> int:
        # Enforce immutability of gap.size/bar.size
        return self._size

    @property
    def position(self) -> str:
        for i in NESW:
            if getattr(self.screen, i) is self:
                return i
        assert False, "Not reached"

    def adjust_reserved_space(self, size: int) -> None:
        for i, side in enumerate(NESW):
            if getattr(self.screen, side) is self:
                self.margin[i] += size
            if self.margin[i] < 0:
                raise ValueError("Gap/Bar can't reserve negative space.")

    @expose_command()
    def info(self) -> dict[str, Any]:
        """
        Info for this object.
        """
        return dict(position=self.position)


class Obj:
    def __init__(self, name: str) -> None:
        self.name = name

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return self.name


STRETCH = Obj("STRETCH")
CALCULATED = Obj("CALCULATED")
STATIC = Obj("STATIC")


class Bar(Gap, configurable.Configurable, CommandObject):
    """A bar, which can contain widgets

    Parameters
    ==========
    widgets :
        A list of widget objects.
    size :
        The "thickness" of the bar, i.e. the height of a horizontal bar, or the
        width of a vertical bar.
    """

    defaults = [
        ("background", "#000000", "Background colour."),
        ("opacity", 1, "Bar window opacity."),
        ("margin", 0, "Space around bar as int or list of ints [N E S W]."),
        ("border_color", "#000000", "Border colour as str or list of str [N E S W]"),
        ("border_width", 0, "Width of border as int of list of ints [N E S W]"),
        (
            "reserve",
            True,
            "Reserve screen space (when set to 'False', bar will be drawn above windows).",
        ),
    ]

    def __init__(self, widgets: list[_Widget], size: int, **config: Any) -> None:
        Gap.__init__(self, size)
        configurable.Configurable.__init__(self, **config)
        self.add_defaults(Bar.defaults)
        # We need to create a new widget list here as users may have the same list for multiple
        # screens. In that scenario, if we replace the widget with a mirror, it replaces it in every
        # bar as python is referring to the same list.
        self.widgets = widgets.copy()
        self.window: Internal | None = None
        self._configured = False
        self._draw_queued = False
        self.future: asyncio.Handle | None = None

        # The part of the margins that was reserved by clients
        self._reserved_space: list[int] = [0, 0, 0, 0]  # [N, E, S, W]
        self._reserved_space_updated = False

        # Size saved when hiding the bar
        self._saved_size = 0

        # Previous window when the bar grabs the keyboard
        self._saved_focus: WindowType | None = None

        # Track widgets that are receiving input
        self._has_cursor: _Widget | None = None
        self._has_keyboard: _Widget | None = None

        # Because Gap.__init__ also sets self.margin
        self.margin = config.get("margin", self.margin)

        # Hacky solution that shows limitations of typing Configurable. We want the
        # option to accept `int | list[int]` but the attribute to be `list[int]`.
        self.margin: list[int]
        if isinstance(self.margin, int):  # type: ignore [unreachable]
            self.margin = [self.margin] * 4  # type: ignore [unreachable]

        self.border_width: list[int]
        if isinstance(self.border_width, int):  # type: ignore [unreachable]
            self.border_width = [self.border_width] * 4  # type: ignore [unreachable]

        self.border_color: ColorsType

        # Check if colours are valid but don't convert to rgba here
        if is_valid_colors(self.border_color):
            if not isinstance(self.border_color, list):
                self.border_color = [self.border_color] * 4
        else:
            logger.warning("Invalid border_color specified. Borders will not be displayed.")
            self.border_width = [0, 0, 0, 0]

    def _configure(self, qtile: Qtile, screen: Screen, reconfigure: bool = False) -> None:
        """
        Configure the bar. `reconfigure` is set to True when screen dimensions
        change, forcing a recalculation of the bar's dimensions.
        """
        # We only want to adjust margin sizes once unless there's new space being
        # reserved or we're reconfiguring the bar because the screen has changed
        if not self._configured or self._reserved_space_updated or reconfigure:
            Gap._configure(self, qtile, screen)

            if any(self.margin) or any(self.border_width) or self._reserved_space_updated:
                # Increase the margin size for the border. The border will be drawn
                # in this space so the empty space will just be the margin.
                margin = [b + s for b, s in zip(self.border_width, self._reserved_space)]

                if self.horizontal:
                    self.x += margin[3] - self.border_width[3]
                    self.width -= margin[1] + margin[3]
                    self._length = self.width
                    self._size += margin[0] + margin[2]
                    if screen.top is self:
                        self.y += margin[0] - self.border_width[0]
                    else:
                        self.y -= margin[2] + self.border_width[2]

                else:
                    self.y += margin[0] - self.border_width[0]
                    self.height -= margin[0] + margin[2]
                    self._length = self.height
                    self._size += margin[1] + margin[3]
                    if screen.left is self:
                        self.x += margin[3] - self.border_width[3]
                    else:
                        self.x -= margin[1] + self.border_width[1]

            if screen.bottom is self and not self.reserve:
                self.y -= self.height + self.margin[2]
            elif screen.right is self and not self.reserve:
                self.x -= self.width + self.margin[1]

            self._reserved_space_updated = False

        width = self.width + (self.border_width[1] + self.border_width[3])
        height = self.height + (self.border_width[0] + self.border_width[2])

        if self.window:
            # We get _configure()-ed with an existing window when screens are getting
            # reconfigured but this screen is present both before and after
            self.window.place(self.x, self.y, width, height, 0, None)

        else:
            # Whereas we won't have a window if we're startup up for the first time or
            # the window has been killed by us no longer using the bar's screen

            # X11 only:
            # To preserve correct display of SysTray widget, we need a 24-bit
            # window where the user requests an opaque bar.
            if qtile.core.name == "x11":
                depth = (
                    32
                    if has_transparency(self.background)
                    else qtile.core.conn.default_screen.root_depth
                )

                self.window = qtile.core.create_internal(  # type: ignore [call-arg]
                    self.x, self.y, width, height, depth
                )

            else:
                self.window = qtile.core.create_internal(self.x, self.y, width, height)

            self.window.opacity = self.opacity
            self.window.unhide()

            self.window.process_window_expose = self.draw
            self.window.process_button_click = self.process_button_click
            self.window.process_button_release = self.process_button_release
            self.window.process_pointer_enter = self.process_pointer_enter
            self.window.process_pointer_leave = self.process_pointer_leave
            self.window.process_pointer_motion = self.process_pointer_motion
            self.window.process_key_press = self.process_key_press

        # We create a new drawer even if there's already a window to ensure the
        # drawer is the right size.
        self.drawer = self.window.create_drawer(width, height)
        self.drawer.clear(self.background)

        crashed_widgets: set[_Widget] = set()
        qtile.renamed_widgets = []
        if self._configured:
            for i in self.widgets:
                if not self._configure_widget(i):
                    crashed_widgets.add(i)
        else:
            for idx, i in enumerate(self.widgets):
                # Create a mirror if this widget is already configured but isn't a Mirror
                # We don't do isinstance(i, Mirror) because importing Mirror (at the top)
                # would give a circular import as libqtile.widget.base imports lbqtile.bar
                if i.configured and i.__class__.__name__ != "Mirror":
                    i = i.create_mirror()
                    self.widgets[idx] = i
                if self._configure_widget(i):
                    qtile.register_widget(i)
                else:
                    crashed_widgets.add(i)

        # Alert the user that we've renamed some widgets
        if qtile.renamed_widgets:
            logger.info(
                "The following widgets were renamed in qtile.widgets_map: %s "
                "To bind commands, rename the widget or use lazy.widget[new_name].",
                ", ".join(qtile.renamed_widgets),
            )
            qtile.renamed_widgets.clear()

        hook.subscribe.setgroup(self.set_layer)
        hook.subscribe.startup_complete(self.set_layer)

        self._remove_crashed_widgets(crashed_widgets)
        self.draw()
        self._resize(self._length, self.widgets)
        self._configured = True

    def _configure_widget(self, widget: _Widget) -> bool:
        assert self.qtile is not None

        if widget.supported_backends and (self.qtile.core.name not in widget.supported_backends):
            logger.warning(
                "Widget removed: %s does not support %s.",
                widget.__class__.__name__,
                self.qtile.core.name,
            )
            return False

        try:
            widget._configure(self.qtile, self)

            if self.horizontal:
                widget.offsety = self.border_width[0]
            else:
                widget.offsetx = self.border_width[3]

            widget.configured = True
        except Exception:
            logger.exception(
                "%s widget crashed during _configure with error:", widget.__class__.__name__
            )
            return False

        return True

    def _remove_crashed_widgets(self, crashed_widgets: set[_Widget]) -> None:
        if not crashed_widgets:
            return

        assert self.qtile is not None
        from libqtile.widget.config_error import ConfigErrorWidget

        for i in crashed_widgets:
            index = self.widgets.index(i)

            # Widgets that aren't available on the current backend should not
            # be shown as "crashed" as the behaviour is expected. Only notify
            # for genuine crashes.
            if not i.supported_backends or (self.qtile.core.name in i.supported_backends):
                crash = ConfigErrorWidget(widget=i)
                crash._configure(self.qtile, self)
                if self.horizontal:
                    crash.offsety = self.border_width[0]
                else:
                    crash.offsetx = self.border_width[3]
                self.widgets.insert(index, crash)

            self.widgets.remove(i)

    def _items(self, name: str) -> ItemT:
        if name == "screen" and self.screen is not None:
            return True, []
        elif name == "widget" and self.widgets:
            return False, [w.name for w in self.widgets]
        return None

    def _select(self, name: str, sel: str | int | None) -> CommandObject | None:
        if name == "screen":
            return self.screen
        elif name == "widget":
            for widget in self.widgets:
                if widget.name == sel:
                    return widget
        return None

    def finalize(self) -> None:
        if self.future:
            self.future.cancel()
        self.drawer.finalize()
        if self.window:
            self.window.kill()
            self.window = None
        self.widgets.clear()

    def kill_window(self) -> None:
        """Kill the window when the bar's screen is no longer being used."""
        assert self.qtile is not None

        if self.future:
            self.future.cancel()

        for name, w in self.qtile.widgets_map.copy().items():
            if w in self.widgets:
                w.finalize()
                del self.qtile.widgets_map[name]
        self.drawer.finalize()
        if self.window:
            self.window.kill()
            self.window = None

        # Reset some flags to allow the bar to be reconfigured as needed
        self._configured = False
        self._draw_queued = False

    def _resize(self, length: int, widgets: list[_Widget]) -> None:
        # We want consecutive stretch widgets to split one 'block' of space between them
        stretches = []
        consecutive_stretches: defaultdict[_Widget, list[_Widget]] = defaultdict(list)
        prev_stretch: _Widget | None = None
        for widget in widgets:
            if widget.length_type == STRETCH:
                if prev_stretch:
                    consecutive_stretches[prev_stretch].append(widget)
                else:
                    stretches.append(widget)
                    prev_stretch = widget
            else:
                prev_stretch = None

        if stretches:
            stretchspace = length - sum(i.length for i in widgets if i.length_type != STRETCH)
            stretchspace = max(stretchspace, 0)
            num_stretches = len(stretches)

            if num_stretches == 1:
                stretches[0].length = stretchspace
            else:
                block = 0
                blocks = []
                for i in widgets:
                    if i.length_type != STRETCH:
                        block += i.length
                    elif i in stretches:  # False for consecutive_stretches
                        blocks.append(block)
                        block = 0
                if block:
                    blocks.append(block)
                interval = length // num_stretches

                for idx, i in enumerate(stretches):
                    if idx == 0:
                        i.length = interval - blocks[0] - blocks[1] // 2
                    elif idx == num_stretches - 1:
                        i.length = interval - blocks[-1] - blocks[-2] // 2
                    else:
                        i.length = int(interval - blocks[idx] / 2 - blocks[idx + 1] / 2)
                    stretchspace -= i.length

                stretches[0].length += stretchspace // 2
                stretches[-1].length += stretchspace - stretchspace // 2

            for i, followers in consecutive_stretches.items():
                length = i.length // (len(followers) + 1)
                rem = i.length - length
                i.length = length
                for f in followers:
                    f.length = length
                    rem -= length
                i.length += rem

        if self.horizontal:
            offset = self.border_width[3]
            for i in widgets:
                i.offsetx = offset
                offset += i.length
        else:
            offset = self.border_width[0]
            for i in widgets:
                i.offsety = offset
                offset += i.length

    def get_widget_in_position(self, x: int, y: int) -> _Widget | None:
        if self.horizontal:
            for i in self.widgets:
                if x < i.offsetx + i.length:
                    return i
        else:
            for i in self.widgets:
                if y < i.offsety + i.length:
                    return i
        return None

    def process_button_click(self, x: int, y: int, button: int) -> None:
        assert self.qtile is not None

        # If we're clicking on a bar that's not on the current screen, focus that screen
        if self.screen and self.screen is not self.qtile.current_screen:
            if self.qtile.core.name == "x11" and self.qtile.current_window:
                self.qtile.current_window._grab_click()
            index = self.qtile.screens.index(self.screen)
            self.qtile.focus_screen(index, warp=False)

        widget = self.get_widget_in_position(x, y)
        if widget:
            widget.button_press(
                x - widget.offsetx,
                y - widget.offsety,
                button,
            )

    def process_button_release(self, x: int, y: int, button: int) -> None:
        widget = self.get_widget_in_position(x, y)
        if widget:
            widget.button_release(
                x - widget.offsetx,
                y - widget.offsety,
                button,
            )

    def process_pointer_enter(self, x: int, y: int) -> None:
        widget = self.get_widget_in_position(x, y)
        if widget:
            widget.mouse_enter(
                x - widget.offsetx,
                y - widget.offsety,
            )
        self._has_cursor = widget

    def process_pointer_leave(self, x: int, y: int) -> None:
        if self._has_cursor:
            self._has_cursor.mouse_leave(
                x - self._has_cursor.offsetx,
                y - self._has_cursor.offsety,
            )
            self._has_cursor = None

    def process_pointer_motion(self, x: int, y: int) -> None:
        widget = self.get_widget_in_position(x, y)
        if widget and self._has_cursor and widget is not self._has_cursor:
            self._has_cursor.mouse_leave(
                x - self._has_cursor.offsetx,
                y - self._has_cursor.offsety,
            )
            widget.mouse_enter(
                x - widget.offsetx,
                y - widget.offsety,
            )
        self._has_cursor = widget

    def process_key_press(self, keycode: int) -> None:
        if self._has_keyboard:
            self._has_keyboard.process_key_press(keycode)

    def widget_grab_keyboard(self, widget: _Widget) -> None:
        """
        A widget can call this method to grab the keyboard focus
        and receive keyboard messages. When done,
        widget_ungrab_keyboard() must be called.
        """
        assert self.qtile is not None

        self._has_keyboard = widget
        self._saved_focus = self.qtile.current_window
        if self.window:
            self.window.focus(False)

    def widget_ungrab_keyboard(self) -> None:
        """
        Removes keyboard focus from the widget.
        """
        if self._saved_focus is not None:
            self._saved_focus.focus(False)
        self._has_keyboard = None

    def draw(self) -> None:
        assert self.qtile is not None

        if not self.widgets:
            return  # calling self._actual_draw in this case would cause a NameError.
        if not self._draw_queued:
            # Delay actually drawing the bar until the event loop is idle, and only once
            # even if this method is called multiple times during the same task.
            self.future = self.qtile.call_soon(self._actual_draw)
            self._draw_queued = True

    def _actual_draw(self) -> None:
        self._draw_queued = False
        self._resize(self._length, self.widgets)
        # We draw the border before the widgets
        if any(self.border_width):
            # The border is drawn "outside" of the bar (i.e. not in the space that the
            # widgets occupy) so we need to add the additional space
            width = self.width + self.border_width[1] + self.border_width[3]
            height = self.height + self.border_width[0] + self.border_width[2]

            # line_opts is a list of tuples where each tuple represents the borders
            # in the order N, E, S, W. The border tuple contains two pairs of
            # co-ordinates for the start and end of the border.
            rects = [
                (0, 0, width, self.border_width[0]),
                (
                    width - (self.border_width[1]),
                    self.border_width[0],
                    self.border_width[1],
                    height - self.border_width[0] - self.border_width[2],
                ),
                (0, height - self.border_width[2], width, self.border_width[2]),
                (
                    0,
                    self.border_width[0],
                    self.border_width[3],
                    height - self.border_width[0] - self.border_width[2],
                ),
            ]

            for border_width, colour, rect in zip(self.border_width, self.border_color, rects):
                if not border_width:
                    continue

                # Draw the border
                self.drawer.clear_rect(*rect)
                self.drawer.ctx.rectangle(*rect)
                self.drawer.set_source_rgb(colour)  # type: ignore[arg-type]
                self.drawer.ctx.fill()
                src_x, src_y, width, height = rect

                self.drawer.draw(
                    offsetx=src_x,
                    offsety=src_y,
                    width=width,
                    height=height,
                    src_x=src_x,
                    src_y=src_y,
                )

        for i in self.widgets:
            i.draw()

        # We need to check if there is any unoccupied space in the bar
        # This can happen where there are no SPACER-type widgets to fill
        # empty space.
        # In that scenario, we fill the empty space with the bar background colour
        # We do this, instead of just filling the bar completely at the start of this
        # method to avoid flickering.

        # Widgets are offset by the top/left border but this is not included in self._length
        # so we adjust the end of the bar area for this offset
        if self.horizontal:
            bar_end = self._length + self.border_width[3]
        else:
            bar_end = self._length + self.border_width[0]

        widget_end = i.offset + i.length

        if widget_end < bar_end:
            # Defines a rectangle for the area enclosed by the bar's borders and the end of the
            # last widget.
            if self.horizontal:
                rect = (widget_end, self.border_width[0], bar_end - widget_end, self.height)
            else:
                rect = (self.border_width[3], widget_end, self.width, bar_end - widget_end)

            # Clear that area (i.e. don't clear borders) and fill with background colour
            self.drawer.clear_rect(*rect)
            self.drawer.ctx.rectangle(*rect)
            self.drawer.set_source_rgb(self.background)
            self.drawer.ctx.fill()
            x, y, w, h = rect
            self.drawer.draw(offsetx=x, offsety=y, height=h, width=w, src_x=x, src_y=y)

    @expose_command()
    def info(self) -> dict[str, Any]:
        return dict(
            size=self._size,
            length=self._length,
            width=self.width,
            height=self.height,
            position=self.position,
            widgets=[i.info() for i in self.widgets],
            window=self.window.wid if self.window else None,
        )

    def is_show(self) -> bool:
        return self._size != 0

    def show(self, is_show: bool = True) -> None:
        if is_show != self.is_show():
            if is_show:
                self._size = self._saved_size
                if self.window:
                    self.window.unhide()
            else:
                self._saved_size = self._size
                self._size = 0
                if self.window:
                    self.window.hide()
            if self.screen and self.screen.group:
                self.screen.group.layout_all()

    def adjust_reserved_space(self, size: int) -> None:
        if self._size:
            # is this necessary?
            self._size = self._initial_size

        for i, side in enumerate(NESW):
            if getattr(self.screen, side) is self:
                self._reserved_space[i] += size
            if self._reserved_space[i] < 0:
                raise ValueError("Gap/Bar can't reserve negative space.")

        self._reserved_space_updated = True

    @expose_command()
    def fake_button_press(
        self, screen: int, position: str, x: int, y: int, button: int = 1
    ) -> None:
        """
        Fake a mouse-button-press on the bar. Co-ordinates are relative
        to the top-left corner of the bar.

        Parameters
        ==========
        widgets :
            A list of widget objects.
        size :
            The "thickness" of the bar, i.e. the height of a horizontal bar, or the
            width of a vertical bar.
        """
        # TODO: drop the screen and position args, update relevant tests
        self.process_button_click(x, y, button)

    def set_layer(self) -> None:
        if self.window:
            if self.reserve:
                self.window.keep_below(enable=True)
            else:
                # Bar is not reserving screen space so let's keep above other windows
                self.window.keep_above(enable=True)


BarType = typing.Union[Bar, Gap]



# Copyright (c) 2011 Florian Mounier
# Copyright (c) 2012, 2014-2015 Tycho Andersen
# Copyright (c) 2013 Mattias Svala
# Copyright (c) 2013 Craig Barnes
# Copyright (c) 2014 ramnes
# Copyright (c) 2014 Sean Vig
# Copyright (c) 2014 Adi Sieker
# Copyright (c) 2014 Chris Wesseling
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

import libqtile.config
from libqtile import layout
from libqtile.confreader import Config
from test.layouts.layout_utils import assert_dimensions, assert_focus_path, assert_focused


class VerticalTileConfig(Config):
    auto_fullscreen = True
    groups = [
        libqtile.config.Group("a"),
        libqtile.config.Group("b"),
        libqtile.config.Group("c"),
        libqtile.config.Group("d"),
    ]
    # default border and margin
    layouts = [layout.VerticalTile(columns=2)]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    screens = []


class VerticalTileSingleBorderConfig(VerticalTileConfig):
    layouts = [layout.VerticalTile(columns=2, single_border_width=2, border_width=8)]


class VerticalTileSingleMarginConfig(VerticalTileConfig):
    layouts = [layout.VerticalTile(columns=2, single_margin=2, margin=8)]


verticaltile_config = pytest.mark.parametrize("manager", [VerticalTileConfig], indirect=True)
verticaltile_single_border_config = pytest.mark.parametrize(
    "manager", [VerticalTileSingleBorderConfig], indirect=True
)
verticaltile_single_margin_config = pytest.mark.parametrize(
    "manager", [VerticalTileSingleMarginConfig], indirect=True
)


@verticaltile_config
def test_verticaltile_simple(manager):
    manager.test_window("one")
    assert_dimensions(manager, 0, 0, 798, 598)
    manager.test_window("two")
    assert_dimensions(manager, 0, 300, 798, 298)
    manager.test_window("three")
    assert_dimensions(manager, 0, 400, 798, 198)


@verticaltile_config
def test_verticaltile_maximize(manager):
    manager.test_window("one")
    assert_dimensions(manager, 0, 0, 798, 598)
    manager.test_window("two")
    assert_dimensions(manager, 0, 300, 798, 298)
    # Maximize the bottom layout, taking 75% of space
    manager.c.layout.maximize()
    assert_dimensions(manager, 0, 150, 798, 448)


@verticaltile_config
def test_verticaltile_window_focus_cycle(manager):
    # setup 3 tiled and two floating clients
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("float1")
    manager.c.window.toggle_floating()
    manager.test_window("float2")
    manager.c.window.toggle_floating()
    manager.test_window("three")

    # test preconditions
    assert manager.c.layout.info()["clients"] == ["one", "two", "three"]
    # last added window has focus
    assert_focused(manager, "three")

    # assert window focus cycle, according to order in layout
    assert_focus_path(manager, "float1", "float2", "one", "two", "three")


@verticaltile_single_border_config
def test_verticaltile_single_border(manager):
    manager.test_window("one")
    assert_dimensions(manager, 0, 0, 796, 596)
    manager.test_window("two")
    assert_dimensions(manager, 0, 300, 784, 284)
    manager.test_window("three")
    assert_dimensions(manager, 0, 400, 784, 184)


@verticaltile_single_margin_config
def test_verticaltile_single_margin(manager):
    manager.test_window("one")
    assert_dimensions(manager, 2, 2, 794, 594)
    manager.test_window("two")
    assert_dimensions(manager, 8, 308, 782, 282)
    manager.test_window("three")
    assert_dimensions(manager, 8, 408, 782, 182)



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging

import pytest

from libqtile.extension.command_set import CommandSet
from libqtile.log_utils import init_log, logger


@pytest.fixture
def fake_qtile():
    class FakeQtile:
        def spawn(self, value):
            logger.warning(value)

    yield FakeQtile()


@pytest.fixture
def log_extension_output(monkeypatch):
    init_log()

    def fake_popen(cmd, *args, **kwargs):
        class PopenObj:
            def communicate(self, value_in, *args):
                if value_in.startswith(b"missing"):
                    return [b"something_else", None]
                else:
                    return [value_in, None]

        return PopenObj()

    monkeypatch.setattr("libqtile.extension.base.Popen", fake_popen)

    yield


@pytest.mark.usefixtures("log_extension_output")
def test_command_set_valid_command(caplog, fake_qtile):
    """Extension should run pre-commands and selected command."""
    extension = CommandSet(pre_commands=["run pre-command"], commands={"key": "run testcommand"})
    extension._configure(fake_qtile)
    extension.run()

    assert caplog.record_tuples == [
        ("libqtile", logging.WARNING, "run pre-command"),
        ("libqtile", logging.WARNING, "run testcommand"),
    ]


@pytest.mark.usefixtures("log_extension_output")
def test_command_set_invalid_command(caplog, fake_qtile):
    """Where the key is not in "commands", no command will be run."""
    extension = CommandSet(
        pre_commands=["run pre-command"], commands={"missing": "run testcommand"}
    )
    extension._configure(fake_qtile)
    extension.run()

    assert caplog.record_tuples == [("libqtile", logging.WARNING, "run pre-command")]


@pytest.mark.usefixtures("log_extension_output")
def test_command_set_inside_command_set_valid_command(caplog, fake_qtile):
    """Extension should run pre-commands and selected command."""

    inside_command = CommandSet(
        pre_commands=["run inner pre-command"],
        commands={"key": "run testcommand"},
    )
    inside_command._configure(fake_qtile)

    extension = CommandSet(
        pre_commands=["run pre-command"],
        commands={"key": inside_command},
    )
    extension._configure(fake_qtile)

    extension.run()

    assert caplog.record_tuples == [
        ("libqtile", logging.WARNING, "run pre-command"),
        (
            "libqtile",
            logging.WARNING,
            "run inner pre-command",
        ),  # pre-command of the inside_command
        ("libqtile", logging.WARNING, "run testcommand"),
    ]


@pytest.mark.usefixtures("log_extension_output")
def test_command_set_inside_command_set_invalid_command(caplog, fake_qtile):
    """Where the key is not in "commands", no command will be run."""
    inside_command = CommandSet(
        pre_commands=["run inner pre-command"],
        commands={"key": "run testcommand"},  # doesn't really matter what command
    )
    inside_command._configure(fake_qtile)

    extension = CommandSet(pre_commands=["run pre-command"], commands={"missing": inside_command})
    extension._configure(fake_qtile)
    extension.run()

    assert caplog.record_tuples == [("libqtile", logging.WARNING, "run pre-command")]

    caplog.clear()

    inside_command = CommandSet(
        pre_commands=["run inner pre-command"],
        commands={"missing": "run testcommand"},
    )
    inside_command._configure(fake_qtile)

    extension = CommandSet(
        pre_commands=["run pre-command"],
        commands={"key": inside_command},
    )
    extension._configure(fake_qtile)

    extension.run()

    assert caplog.record_tuples == [
        ("libqtile", logging.WARNING, "run pre-command"),
        (
            "libqtile",
            logging.WARNING,
            "run inner pre-command",
        ),  # pre-command of the inside_command
    ]



# Copyright (c) 2023, elParaguayo. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from libqtile.scripts.migrations._base import (
    Check,
    RenamerTransformer,
    _QtileMigrator,
    add_migration,
)


class RenameTileMasterTransformer(RenamerTransformer):
    from_to = ("masterWindows", "master_length")


class RenameTileMaster(_QtileMigrator):
    ID = "RenameTileMaster"

    SUMMARY = "Changes ``masterWindows`` argument to ``master_length``."

    HELP = """
    To be consistent with other layouts, the ``masterWindows`` property of the
    ``Tile`` layout was renamed to ``master_length``. Configs using the
    ``masterWindows`` argument when configuring the layout should replace this.
    """

    AFTER_VERSION = "0.16.1"

    TESTS = [
        Check(
            """
            from libqtile import layout
            from libqtile.layout import Tile

            layouts = [
                Tile(masterWindows=2),
                layout.Tile(masterWindows=3)
            ]
            """,
            """
            from libqtile import layout
            from libqtile.layout import Tile

            layouts = [
                Tile(master_length=2),
                layout.Tile(master_length=3)
            ]
            """,
        )
    ]

    visitor = RenameTileMasterTransformer


add_migration(RenameTileMaster)



# -*- coding: utf-8 -*-
# Copyright (c) 2011 Florian Mounier
# Copyright (c) 2011 Mounier Florian
# Copyright (c) 2012 roger
# Copyright (c) 2012-2014 Tycho Andersen
# Copyright (c) 2012-2013 Craig Barnes
# Copyright (c) 2013 Tao Sauvage
# Copyright (c) 2014 Sean Vig
# Copyright (c) 2014 Adi Sieker
# Copyright (c) 2020 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from os import path

from libqtile import bar, pangocffi, utils
from libqtile.command.base import expose_command
from libqtile.log_utils import logger
from libqtile.notify import ClosedReason, notifier
from libqtile.widget import base


class Notify(base._TextBox):
    """
    A notify widget

    This widget can handle actions provided by notification clients. However, only the
    default action is supported, so if a client provides multiple actions then only the
    default (first) action can be invoked. Some programs will provide their own
    notification windows if the notification server does not support actions, so if you
    want your notifications to handle more than one action then specify ``False`` for
    the ``action`` option to disable all action handling. Unfortunately we cannot
    specify the capability for exactly one action.
    """

    defaults = [
        ("foreground_urgent", "ff0000", "Foreground urgent priority colour"),
        ("foreground_low", "dddddd", "Foreground low priority  colour"),
        ("default_timeout_low", 5, "Default timeout (seconds) for low urgency notifications."),
        ("default_timeout", 10, "Default timeout (seconds) for normal notifications"),
        ("default_timeout_urgent", None, "Default timeout (seconds) for urgent notifications"),
        ("audiofile", None, "Audiofile played during notifications"),
        ("action", True, "Enable handling of default action upon right click"),
        (
            "parse_text",
            None,
            "Function to parse and modify notifications. "
            "e.g. function in config that removes line returns:"
            "def my_func(text)"
            "   return text.replace('\n', '')"
            "then set option parse_text=my_func",
        ),
        ("background_urgent", "440000", "Background urgent priority colour"),
        ("background_low", "444444", "Background low priority colour"),
    ]
    capabilities = {"body", "actions"}

    def __init__(self, width=bar.CALCULATED, **config):
        base._TextBox.__init__(self, "", width, **config)
        self.add_defaults(Notify.defaults)
        self.current_id = 0

        default_callbacks = {
            "Button1": self.clear,
            "Button4": self.prev,
            "Button5": self.next,
        }
        if self.action:
            default_callbacks["Button3"] = self._invoke
        else:
            self.capabilities = Notify.capabilities.difference({"actions"})
        self.add_callbacks(default_callbacks)

        self.background_normal = self.background

    def _configure(self, qtile, bar):
        base._TextBox._configure(self, qtile, bar)
        self.layout = self.drawer.textlayout(
            self.text, self.foreground, self.font, self.fontsize, self.fontshadow, markup=True
        )
        if notifier is None:
            logger.warning("You must install dbus-next to use the Notify widget.")

        # Create a tuple of our default timeouts. Urgency is an integer of 0-2
        # (see https://specifications.freedesktop.org/notification-spec/notification-spec-latest.html#urgency-levels)
        # so they will work as the index of the tuple.
        self._timeouts = (
            self.default_timeout_low,
            self.default_timeout,
            self.default_timeout_urgent,
        )

    async def _config_async(self):
        if notifier is None:
            return

        await notifier.register(self.update, self.capabilities, on_close=self.on_close)

    def set_notif_text(self, notif):
        self.text = pangocffi.markup_escape_text(notif.summary)
        urgency = getattr(notif.hints.get("urgency"), "value", 1)
        if urgency != 1:
            self.text = '<span color="%s">%s</span>' % (
                utils.hex(self.foreground_urgent if urgency == 2 else self.foreground_low),
                self.text,
            )
            self.background = self.background_urgent if urgency == 2 else self.background_low
        else:
            self.background = self.background_normal

        if notif.body:
            self.text = '<span weight="bold">%s</span> - %s' % (
                self.text,
                pangocffi.markup_escape_text(notif.body),
            )
        if callable(self.parse_text):
            try:
                self.text = self.parse_text(self.text)
            except:  # noqa: E722
                logger.exception("parse_text function failed:")
        if self.audiofile and path.exists(self.audiofile):
            self.qtile.spawn("aplay -q '%s'" % self.audiofile)

    def update(self, notif):
        self.qtile.call_soon_threadsafe(self.real_update, notif)

    def real_update(self, notif):
        self.set_notif_text(notif)
        self.current_id = notif.id - 1
        if notif.timeout and notif.timeout > 0:
            self.timeout_add(
                notif.timeout / 1000, self.clear, method_args=(ClosedReason.expired,)
            )
        else:
            urgency = getattr(notif.hints.get("urgency"), "value", 1)
            try:
                timeout = self._timeouts[urgency]
            except IndexError:
                logger.warning(
                    "Notification had an unexpected urgency value. Treating as normal priority."
                )
                timeout = self._timeouts[1]

            if timeout:
                self.timeout_add(timeout, self.clear, method_args=(ClosedReason.expired,))
        self.bar.draw()
        return True

    @expose_command()
    def display(self):
        if notifier is None:
            return

        self.set_notif_text(notifier.notifications[self.current_id])
        self.bar.draw()

    @expose_command()
    def clear(self, reason=ClosedReason.dismissed):
        """Clear the notification"""
        if notifier is None:
            return

        notifier._service.NotificationClosed(notifier.notifications[self.current_id].id, reason)
        self.text = ""
        self.background = self.background_normal
        self.current_id = len(notifier.notifications) - 1
        self.bar.draw()

    def on_close(self, nid):
        if self.current_id < len(notifier.notifications):
            notif = notifier.notifications[self.current_id]
            if notif.id == nid:
                self.clear(ClosedReason.method)

    @expose_command()
    def prev(self):
        """Show previous notification."""
        if self.current_id > 0:
            self.current_id -= 1
        self.display()

    @expose_command()
    def next(self):
        if notifier is None:
            return

        """Show next notification."""
        if self.current_id < len(notifier.notifications) - 1:
            self.current_id += 1
            self.display()

    def _invoke(self):
        if self.current_id < len(notifier.notifications):
            notif = notifier.notifications[self.current_id]
            if notif.actions:
                notifier._service.ActionInvoked(notif.id, notif.actions[0])
            self.clear()

    @expose_command()
    def toggle(self):
        """Toggle showing/clearing the notification"""
        if self.text == "":
            self.display()
        else:
            self.clear()

    @expose_command()
    def invoke(self):
        """Invoke the notification's default action"""
        if self.action:
            self._invoke()

    def finalize(self):
        notifier.unregister(self.update, on_close=self.on_close)
        base._TextBox.finalize(self)



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Widget specific tests

import builtins
import sys
from importlib import reload
from types import ModuleType

import pytest

import libqtile.config
from libqtile.bar import Bar


def no_op(*args, **kwargs):
    pass


def mock_open(output):
    class MockOpen:
        def __init__(self, *args):
            self.output = output

        def __enter__(self):
            return self

        def __exit__(*exc):
            return None

        def read(self):
            return self.output

    return MockOpen


class MockIwlib(ModuleType):
    DATA = {
        "wlan0": {
            "NWID": b"Auto",
            "Frequency": b"5.18 GHz",
            "Access Point": b"12:34:56:78:90:AB",
            "BitRate": b"650 Mb/s",
            "ESSID": b"QtileNet",
            "Mode": b"Managed",
            "stats": {"quality": 49, "level": 190, "noise": 0, "updated": 75},
        },
        "wlan1": {
            "ESSID": None,
        },
    }

    @classmethod
    def get_iwconfig(cls, interface):
        return cls.DATA.get(interface, dict())


# Patch the widget with our mock iwlib module.
@pytest.fixture
def patched_wlan(monkeypatch):
    monkeypatch.setitem(sys.modules, "iwlib", MockIwlib("iwlib"))
    from libqtile.widget import wlan

    # Reload fixes cases where psutil may have been imported previously
    reload(wlan)

    yield wlan


@pytest.mark.parametrize(
    "kwargs,expected",
    [
        ({}, "QtileNet 49/70"),
        ({"format": "{essid} {percent:2.0%}"}, "QtileNet 70%"),
        ({"interface": "wlan1"}, "Disconnected"),
    ],
)
def test_wlan_display(minimal_conf_noscreen, manager_nospawn, patched_wlan, kwargs, expected):
    widget = patched_wlan.Wlan(**kwargs)
    config = minimal_conf_noscreen
    config.screens = [libqtile.config.Screen(top=Bar([widget], 10))]
    manager_nospawn.start(config)

    text = manager_nospawn.c.bar["top"].info()["widgets"][0]["text"]
    assert text == expected


@pytest.mark.parametrize(
    "kwargs,state,expected",
    [
        ({"interface": "wlan1", "use_ethernet": True}, "up", "eth"),
        ({"interface": "wlan1", "use_ethernet": True}, "down", "Disconnected"),
        (
            {"interface": "wlan1", "use_ethernet": True, "ethernet_message": "Wired"},
            "up",
            "Wired",
        ),
    ],
)
def test_ethernet(
    minimal_conf_noscreen, manager_nospawn, patched_wlan, kwargs, state, expected, monkeypatch
):
    monkeypatch.setattr(builtins, "open", mock_open(state))
    widget = patched_wlan.Wlan(**kwargs)
    assert widget.poll() == expected



# Copyright (c) 2024 Florian G. Hechler

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import tempfile
from importlib import reload

import pytest

from test.widgets.test_hdd import MockPsutil, set_io_ticks


@pytest.fixture
def widget(monkeypatch):
    monkeypatch.setitem(sys.modules, "psutil", MockPsutil("psutil"))
    from libqtile.widget import hdd

    reload(hdd)
    yield hdd.HDD


def ss_cpu(screenshot_manager):
    # Create a fake stat file
    temp_file = tempfile.NamedTemporaryFile(mode="w+", delete=False)
    widget = screenshot_manager.c.widget["hdd"]
    widget.eval(f"self.path = '{temp_file.name}'")

    set_io_ticks(temp_file, 0)
    widget.eval("self.update(self.poll())")

    set_io_ticks(temp_file, 123000)
    widget.eval("self.update(self.poll())")

    screenshot_manager.take_screenshot()



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Widget specific tests

from os import path

import pytest

import libqtile.bar
import libqtile.config
from libqtile import widget

TEST_DIR = path.dirname(path.abspath(__file__))
DATA_DIR = path.join(TEST_DIR, "..", "data", "png")
IMAGE_FILE = path.join(DATA_DIR, "audio-volume-muted.png")


img = widget.Image(filename=IMAGE_FILE)

parameters = [
    (libqtile.config.Screen(top=libqtile.bar.Bar([img], 40)), "top", "height"),
    (libqtile.config.Screen(left=libqtile.bar.Bar([img], 40)), "left", "width"),
]


@pytest.mark.parametrize("screen,location,attribute", parameters)
def test_default_settings(manager_nospawn, minimal_conf_noscreen, screen, location, attribute):
    config = minimal_conf_noscreen
    config.screens = [screen]

    manager_nospawn.start(config)
    bar = manager_nospawn.c.bar[location]

    info = bar.info()
    for dimension in ["height", "width"]:
        assert info["widgets"][0][dimension] == info[attribute]


no_img = widget.Image()

parameters = [
    (libqtile.config.Screen(top=libqtile.bar.Bar([no_img], 40)), "top", "width"),
    (libqtile.config.Screen(left=libqtile.bar.Bar([no_img], 40)), "left", "height"),
]


@pytest.mark.parametrize("screen,location,attribute", parameters)
def test_no_filename(manager_nospawn, minimal_conf_noscreen, screen, location, attribute):
    config = minimal_conf_noscreen
    config.screens = [screen]

    manager_nospawn.start(config)
    bar = manager_nospawn.c.bar[location]

    info = bar.info()
    assert info["widgets"][0][attribute] == 0


def test_missing_file(manager_nospawn, minimal_conf_noscreen):
    img2 = widget.Image(filename="/this/file/does/not/exist")

    config = minimal_conf_noscreen
    config.screens = [libqtile.config.Screen(top=libqtile.bar.Bar([img2], 40))]

    manager_nospawn.start(config)
    bar = manager_nospawn.c.bar["top"]

    info = bar.info()
    assert info["widgets"][0]["width"] == 0


def test_no_scale(manager_nospawn, minimal_conf_noscreen):
    img2 = widget.Image(filename=IMAGE_FILE, scale=False)

    config = minimal_conf_noscreen
    config.screens = [libqtile.config.Screen(top=libqtile.bar.Bar([img2], 40))]

    manager_nospawn.start(config)
    bar = manager_nospawn.c.bar["top"]

    info = bar.info()
    assert info["widgets"][0]["width"] == 24


def test_no_image(manager_nospawn, minimal_conf_noscreen, logger):
    img = widget.Image()

    config = minimal_conf_noscreen
    config.screens = [libqtile.config.Screen(top=libqtile.bar.Bar([img], 40))]

    manager_nospawn.start(config)

    assert "Image filename not set!" in logger.text


def test_invalid_path(manager_nospawn, minimal_conf_noscreen, logger):
    filename = "/made/up/file.png"
    img = widget.Image(filename=filename)

    config = minimal_conf_noscreen
    config.screens = [libqtile.config.Screen(top=libqtile.bar.Bar([img], 40))]

    manager_nospawn.start(config)

    assert f"Image does not exist: {filename}" in logger.text



# Copyright (c) 2024 Florian G. Hechler

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from libqtile.widget import base


class HDD(base.ThreadPoolText):
    """
    Displays HDD usage in percent based on the number of milliseconds the device has been performing I/O operations.
    """

    defaults = [
        ("device", "sda", "Block device to monitor (e.g. sda)"),
        (
            "format",
            "HDD {HDDPercent}%",
            "HDD display format",
        ),
    ]

    def __init__(self, **config):
        super().__init__("", **config)
        self.add_defaults(HDD.defaults)
        self.path = "/sys/block/{dev}/stat".format(dev=self.device)
        self._prev = 0

    def poll(self):
        variables = dict()
        # Field index 9 contains the number of milliseconds the device has been performing I/O operations
        with open(self.path) as f:
            io_ticks = int(f.read().split()[9])

        variables["HDDPercent"] = round(
            max(min(((io_ticks - self._prev) / self.update_interval) / 10, 100.0), 0.0), 1
        )

        self._prev = io_ticks

        return self.format.format(**variables)



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from libqtile import widget

RESPONSE = {
    "Meta Data": {
        "1. Information": "Intraday (1min) open, high, low, close prices and volume",
        "2. Symbol": "QTIL",
        "3. Last Refreshed": "2021-07-30 19:09:00",
        "4. Interval": "1min",
        "5. Output Size": "Compact",
        "6. Time Zone": "US/Eastern",
    },
    "Time Series (1min)": {
        "2021-07-30 19:09:00": {
            "1. open": "140.9800",
            "2. high": "140.9800",
            "3. low": "140.9800",
            "4. close": "140.9800",
            "5. volume": "527",
        },
        "2021-07-30 17:27:00": {
            "1. open": "141.1900",
            "2. high": "141.1900",
            "3. low": "141.1900",
            "4. close": "141.1900",
            "5. volume": "300",
        },
        "2021-07-30 16:44:00": {
            "1. open": "141.0000",
            "2. high": "141.0000",
            "3. low": "141.0000",
            "4. close": "141.0000",
            "5. volume": "482",
        },
        "2021-07-30 16:26:00": {
            "1. open": "141.0000",
            "2. high": "141.0000",
            "3. low": "141.0000",
            "4. close": "141.0000",
            "5. volume": "102",
        },
    },
}


def test_stock_ticker_methods():
    ticker = widget.StockTicker(symbol="QTIL")

    assert ticker.url == (
        "https://www.alphavantage.co/query?interval=1min&outputsize=compact&"
        "function=TIME_SERIES_INTRADAY&symbol=QTIL"
    )

    # We don't know what locale is on the testing system but we can just use
    # whatever the widget is using.
    assert ticker.parse(RESPONSE) == f"QTIL: {ticker.sign}140.98"



import pytest
import xcffib
import xcffib.testing

from libqtile.backend.x11 import window, xcbq


def test_new_window(conn):
    win = conn.create_window(1, 2, 640, 480)
    assert isinstance(win, window.XWindow)
    geom = win.get_geometry()
    assert geom.x == 1
    assert geom.y == 2
    assert geom.width == 640
    assert geom.height == 480
    win.kill_client()
    with pytest.raises(xcffib.ConnectionException):
        win.get_geometry()


def test_masks():
    cfgmasks = xcbq.ConfigureMasks
    d = {"x": 1, "y": 2, "width": 640, "height": 480}
    mask, vals = cfgmasks(**d)
    assert set(vals) == set(d.values())
    with pytest.raises(ValueError):
        mask, vals = cfgmasks(asdf=32, **d)


def test_translate_masks():
    assert xcbq.translate_masks(["shift", "control"])
    assert xcbq.translate_masks([]) == 0



# Copyright (c) 2021 Matt Colligan
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import os
from typing import TYPE_CHECKING

from wlroots.util.box import Box
from wlroots.util.clock import Timespec
from wlroots.wlr_types import Output as wlrOutput
from wlroots.wlr_types import OutputState, SceneOutput
from wlroots.wlr_types.layer_shell_v1 import (
    LayerShellV1Layer,
    LayerSurfaceV1KeyboardInteractivity,
)
from wlroots.wlr_types.output import CustomMode

from libqtile.backend.wayland.wlrq import HasListeners
from libqtile.config import ScreenRect
from libqtile.log_utils import logger

if TYPE_CHECKING:
    from typing import Any

    from pywayland.server import Listener
    from wlroots.wlr_types.output import OutputEventRequestState

    from libqtile.backend.wayland.core import Core
    from libqtile.backend.wayland.layer import LayerStatic
    from libqtile.backend.wayland.window import WindowType
    from libqtile.config import Screen


class Output(HasListeners):
    def __init__(self, core: Core, wlr_output: wlrOutput):
        self.core = core
        self.renderer = core.renderer
        self.wlr_output = wlr_output
        self._reserved_space = (0, 0, 0, 0)

        # These will get updated on the output layout's change event
        self.x = 0
        self.y = 0

        self.scene_output = SceneOutput.create(core.scene, wlr_output)
        wlr_output.init_render(core.allocator, core.renderer)

        # The output may be disabled, switch it on.
        state = OutputState()
        state.set_enabled(True)

        # Select the output's preferred mode.
        if mode := wlr_output.preferred_mode():
            state.set_mode(mode)

        # During tests, we want to fix the geometry of the 1 or 2 outputs.
        if wlr_output.is_headless and "PYTEST_CURRENT_TEST" in os.environ:
            if not core.get_enabled_outputs():
                # First test output
                state.set_custom_mode(CustomMode(width=800, height=600, refresh=0))
            else:
                # Second test output
                state.set_custom_mode(CustomMode(width=640, height=480, refresh=0))

        # Commit this initial state.
        wlr_output.commit(state)
        state.finish()

        wlr_output.data = self

        self.add_listener(wlr_output.destroy_event, self._on_destroy)
        self.add_listener(wlr_output.frame_event, self._on_frame)
        self.add_listener(wlr_output.request_state_event, self._on_request_state)

        # The layers enum indexes into this list to get a list of surfaces
        self.layers: list[list[LayerStatic]] = [[] for _ in range(len(LayerShellV1Layer))]

    def finalize(self) -> None:
        self.finalize_listeners()
        self.core.remove_output(self)
        self.scene_output.destroy()

    def __repr__(self) -> str:
        i = self.get_screen_info()
        return f"<Output ({i.x}, {i.y}, {i.width}, {i.height})>"

    @property
    def screen(self) -> Screen:
        assert self.core.qtile is not None

        for screen in self.core.qtile.screens:
            # Outputs alias if they have the same (x, y) and share the same Screen, so
            # we don't need to check the if the width and height match the Screen's.
            if screen.x == self.x and screen.y == self.y:
                return screen

        return self.core.qtile.current_screen

    def _on_destroy(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: output destroy")
        self.finalize()

    def _on_frame(self, _listener: Listener, _data: Any) -> None:
        self.core.configure_node_opacity(self.core.windows_tree.node)
        try:
            self.scene_output.commit()
        except RuntimeError:
            # Failed to commit scene output; skip rendering.
            pass

        # Inform clients of the frame
        self.scene_output.send_frame_done(Timespec.get_monotonic_time())

    def _on_request_state(self, _listener: Listener, request: OutputEventRequestState) -> None:
        logger.debug("Signal: output request_state")
        self.wlr_output.commit(request.state)

    def get_screen_info(self) -> ScreenRect:
        width, height = self.wlr_output.effective_resolution()
        return ScreenRect(int(self.x), int(self.y), width, height)

    def organise_layers(self) -> None:
        """Organise the positioning of layer shell surfaces."""
        logger.debug("Output: organising layers")
        ow, oh = self.wlr_output.effective_resolution()

        # These rects are in output layout coordinates
        full_area = Box(self.x, self.y, ow, oh)
        usable_area = Box(self.x, self.y, ow, oh)

        for layer in reversed(LayerShellV1Layer):
            # Arrange exclusive surface from top to bottom
            self._organise_layer(layer, full_area, usable_area, exclusive=True)

        # TODO: can this be a geometry?
        # The positions used for reserving space are screen-relative coordinates
        new_reserved_space = (
            usable_area.x - self.x,  # left
            self.x + ow - usable_area.x - usable_area.width,  # right
            usable_area.y - self.y,  # top
            self.y + oh - usable_area.y - usable_area.height,  # bottom
        )
        delta = tuple(new - old for new, old in zip(new_reserved_space, self._reserved_space))
        if any(delta):
            self.core.qtile.reserve_space(delta, self.screen)  # type: ignore
            self._reserved_space = new_reserved_space

        for layer in reversed(LayerShellV1Layer):
            # Arrange non-exclusive surface from top to bottom
            self._organise_layer(layer, full_area, usable_area, exclusive=False)

        # Find topmost keyboard interactive layer
        for layer in (LayerShellV1Layer.OVERLAY, LayerShellV1Layer.TOP):
            for win in self.layers[layer]:
                if (
                    win.surface.current.keyboard_interactive
                    == LayerSurfaceV1KeyboardInteractivity.EXCLUSIVE
                ):
                    self.core.exclusive_layer = win
                    self.core.focus_window(win)
                    return
                if self.core.exclusive_layer is win:
                    # This window previously had exclusive focus, but no longer wants it.
                    self.core.exclusive_layer = None

    def _organise_layer(
        self,
        layer: LayerShellV1Layer,
        full_area: Box,
        usable_area: Box,
        *,
        exclusive: bool,
    ) -> None:
        for win in self.layers[layer]:
            state = win.surface.current

            if exclusive != (0 < state.exclusive_zone):
                continue

            win.scene_layer.configure(full_area, usable_area)
            win.place(
                win.tree.node.x,
                win.tree.node.y,
                state.desired_width,
                state.desired_height,
                0,
                None,
            )

    def contains(self, rect: WindowType) -> bool:
        """Returns whether the given window is visible on this output."""
        if rect.x + rect.width < self.x:
            return False
        if rect.y + rect.height < self.y:
            return False

        ow, oh = self.wlr_output.effective_resolution()
        if self.x + ow < rect.x:
            return False
        if self.y + oh < rect.y:
            return False

        return True



# This is used for test_manager.py::test_reload_config
#
# The exported configuration variables have a different value depending on whether
# libqtile has a 'test_data' attribute (see below)

import sys
from pathlib import Path

from libqtile import bar, layout, qtile, widget
from libqtile.config import Drag, DropDown, Group, Key, Match, Rule, ScratchPad, Screen
from libqtile.dgroups import simple_key_binder
from libqtile.lazy import lazy

keys = [
    Key(["mod4"], "h", lazy.layout.left(), desc="Move focus to left"),
]

groups = [Group(i) for i in "12345"]

layouts = [
    layout.Columns(border_focus_stack=["#d75f5f", "#8f3d3d"], border_width=4),
]

screens = [
    Screen(
        bottom=bar.Bar(
            [
                widget.CurrentLayout(),
                widget.GroupBox(),
                widget.Clock(format="%Y-%m-%d %a %I:%M %p"),
                widget.QuickExit(),
            ],
            24,
        ),
    ),
]

widget_defaults = dict()

mouse = [
    Drag(
        ["mod4"],
        "Button1",
        lazy.window.set_position_floating(),
        start=lazy.window.get_position(),
    ),
]

windowpy = Path(__file__).parent.parent / "scripts" / "window.py"
script = " ".join([sys.executable, windowpy.as_posix(), "--name", "dd", "dd", "normal"])
dropdowns = [DropDown("dropdown1", script)]

dgroups_key_binder = None
dgroups_app_rules = []
floating_layout = layout.Floating(float_rules=[Match(title="one")])
wmname = "LG3D"


if hasattr(qtile, "test_data"):
    # Add more items or change values qtile.test_data is set
    keys.append(
        Key(["mod4"], "l", lazy.layout.right(), desc="Move focus to right"),
    )

    groups.extend([Group(i) for i in "6789"])

    layouts.append(layout.Max())

    screens = [
        Screen(top=bar.Bar([widget.CurrentLayout()], 32)),
    ]

    widget_defaults["background"] = "#ff0000"

    mouse.append(
        Drag(
            ["mod4"],
            "Button3",
            lazy.window.set_size_floating(),
            start=lazy.window.get_size(),
        ),
    )

    dropdowns.append(DropDown("dropdown2", script))

    dgroups_key_binder = simple_key_binder
    dgroups_app_rules = [Rule(Match(wm_class="test"), float=True)]
    floating_layout = layout.Floating()
    wmname = "TEST"

    qtile.test_data_config_evaluations += 1

groups.append(ScratchPad("S", dropdowns))



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import pytest

import libqtile.bar
import libqtile.config
import libqtile.layout
from libqtile.confreader import Config
from libqtile.extension.window_list import WindowList
from libqtile.lazy import lazy


@pytest.fixture
def extension_manager(monkeypatch, manager_nospawn):
    extension = WindowList()

    # We want the value returned immediately
    def fake_popen(cmd, *args, **kwargs):
        class PopenObj:
            def communicate(self, value_in, *args):
                return [value_in, None]

        return PopenObj()

    monkeypatch.setattr("libqtile.extension.base.Popen", fake_popen)

    class ManagerConfig(Config):
        groups = [
            libqtile.config.Group("a"),
            libqtile.config.Group("b"),
        ]
        layouts = [libqtile.layout.max.Max()]
        keys = [
            libqtile.config.Key(["control"], "k", lazy.run_extension(extension)),
        ]
        screens = [
            libqtile.config.Screen(
                bottom=libqtile.bar.Bar([], 20),
            )
        ]

    manager_nospawn.start(ManagerConfig)

    yield manager_nospawn


def test_window_list(extension_manager):
    """Test WindowList extension switches group."""

    # Launch a window and verify it's on the current group
    extension_manager.test_window("one")
    assert len(extension_manager.c.group.info()["windows"]) == 1

    # Switch group and verify no windows in group
    extension_manager.c.group["b"].toscreen()
    assert len(extension_manager.c.group.info()["windows"]) == 0

    # Toggle extension (which is patched to return immediately)
    # Check that window is visible on original group
    extension_manager.c.simulate_keypress(["control"], "k")
    assert len(extension_manager.c.group.info()["windows"]) == 1
    assert extension_manager.c.group.info()["label"] == "a"



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Widget specific tests

import pytest

import libqtile.config
from libqtile import bar, layout, widget
from libqtile.config import Screen
from libqtile.confreader import Config


class WindowNameConfig(Config):
    auto_fullscreen = True
    groups = [libqtile.config.Group("a"), libqtile.config.Group("b")]
    layouts = [layout.Max()]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    fake_screens = [
        Screen(
            top=bar.Bar(
                [
                    widget.WindowName(),
                ],
                24,
            ),
            x=0,
            y=0,
            width=900,
            height=480,
        ),
        Screen(
            top=bar.Bar(
                [
                    widget.WindowName(for_current_screen=True),
                ],
                24,
            ),
            x=0,
            y=480,
            width=900,
            height=480,
        ),
    ]
    screens = []


windowname_config = pytest.mark.parametrize("manager", [WindowNameConfig], indirect=True)


@windowname_config
def test_window_names(manager):
    def widget_text_on_screen(index):
        return manager.c.screen[index].bar["top"].info()["widgets"][0]["text"]

    # Screen 1's widget is set up with for_current_screen=True
    # This means that when screen 0 is active, screen 1's widget should show the same text

    assert widget_text_on_screen(0) == " "
    assert widget_text_on_screen(0) == widget_text_on_screen(1)

    # Load a window
    proc = manager.test_window("one")
    assert widget_text_on_screen(0) == "one"
    assert widget_text_on_screen(0) == widget_text_on_screen(1)

    # Maximize window
    manager.c.window.toggle_maximize()
    assert widget_text_on_screen(0) == "[] one"
    assert widget_text_on_screen(0) == widget_text_on_screen(1)

    # Minimize window
    manager.c.window.toggle_minimize()
    assert widget_text_on_screen(0) == "_ one"
    assert widget_text_on_screen(0) == widget_text_on_screen(1)

    # Float window
    manager.c.window.toggle_minimize()
    manager.c.window.toggle_floating()
    assert widget_text_on_screen(0) == "V one"
    assert widget_text_on_screen(0) == widget_text_on_screen(1)

    # Kill the window and check text again
    manager.kill_window(proc)
    assert widget_text_on_screen(0) == " "
    assert widget_text_on_screen(0) == widget_text_on_screen(1)

    # Quick test to check for_current_screen=False works
    manager.c.to_screen(1)
    proc = manager.test_window("one")
    assert widget_text_on_screen(0) == " "
    assert widget_text_on_screen(1) == "one"
    manager.kill_window(proc)



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Widget specific tests

import subprocess

import pytest

import libqtile.config
from libqtile.widget import moc
from test.widgets.conftest import FakeBar


class MockMocpProcess:
    info = {}
    is_error = False
    index = 0

    @classmethod
    def reset(cls):
        cls.info = [
            {
                "State": "PLAY",
                "File": "/playing/file/rickroll.mp3",
                "SongTitle": "Never Gonna Give You Up",
                "Artist": "Rick Astley",
                "Album": "Whenever You Need Somebody",
            },
            {
                "State": "PLAY",
                "File": "/playing/file/sweetcaroline.mp3",
                "SongTitle": "Sweet Caroline",
                "Artist": "Neil Diamond",
                "Album": "Greatest Hits",
            },
            {
                "State": "STOP",
                "File": "/playing/file/itsnotunusual.mp3",
                "SongTitle": "It's Not Unusual",
                "Artist": "Tom Jones",
                "Album": "Along Came Jones",
            },
        ]
        cls.index = 0

    @classmethod
    def run(cls, cmd):
        if cls.is_error:
            raise subprocess.CalledProcessError(-1, cmd=cmd, output="Couldn't connect to moc.")

        arg = cmd[1]

        if arg == "-i":
            output = "\n".join(
                "{k}: {v}".format(k=k, v=v) for k, v in cls.info[cls.index].items()
            )
            return output

        elif arg == "-p":
            cls.info[cls.index]["State"] = "PLAY"

        elif arg == "-G":
            if cls.info[cls.index]["State"] == "PLAY":
                cls.info[cls.index]["State"] = "PAUSE"

            elif cls.info[cls.index]["State"] == "PAUSE":
                cls.info[cls.index]["State"] = "PLAY"

        elif arg == "-f":
            cls.index = (cls.index + 1) % len(cls.info)

        elif arg == "-r":
            cls.index = (cls.index - 1) % len(cls.info)


def no_op(*args, **kwargs):
    pass


@pytest.fixture
def patched_moc(fake_qtile, monkeypatch, fake_window):
    widget = moc.Moc()
    MockMocpProcess.reset()
    monkeypatch.setattr(widget, "call_process", MockMocpProcess.run)
    monkeypatch.setattr("libqtile.widget.moc.subprocess.Popen", MockMocpProcess.run)
    fakebar = FakeBar([widget], window=fake_window)
    widget._configure(fake_qtile, fakebar)
    return widget


def test_moc_poll_string_formatting(patched_moc):
    # Both artist and song title
    assert patched_moc.poll() == "♫ Rick Astley - Never Gonna Give You Up"

    # No artist
    MockMocpProcess.info[0]["Artist"] = ""
    assert patched_moc.poll() == "♫ Never Gonna Give You Up"

    # No title
    MockMocpProcess.info[0]["SongTitle"] = ""
    assert patched_moc.poll() == "♫ rickroll"


def test_moc_state_and_colours(patched_moc):
    # Initial poll - playing
    patched_moc.poll()
    assert patched_moc.layout.colour == patched_moc.play_color

    # Toggle pause
    patched_moc.play()
    patched_moc.poll()
    assert patched_moc.layout.colour == patched_moc.noplay_color

    # Toggle pause --> playing again
    patched_moc.play()
    patched_moc.poll()
    assert patched_moc.layout.colour == patched_moc.play_color


def test_moc_button_presses(manager_nospawn, minimal_conf_noscreen, monkeypatch):
    # This needs to be patched before initialising the widgets as mouse callbacks
    # bind subprocess.Popen.
    monkeypatch.setattr("subprocess.Popen", MockMocpProcess.run)

    # Long interval as we don't need this polling on its own.
    mocwidget = moc.Moc(update_interval=30)
    MockMocpProcess.reset()
    monkeypatch.setattr(mocwidget, "call_process", MockMocpProcess.run)
    monkeypatch.setattr("libqtile.widget.moc.subprocess.Popen", MockMocpProcess.run)

    config = minimal_conf_noscreen
    config.screens = [libqtile.config.Screen(top=libqtile.bar.Bar([mocwidget], 10))]

    manager_nospawn.start(config)

    # When started, we have the first item playing
    topbar = manager_nospawn.c.bar["top"]
    info = manager_nospawn.c.widget["moc"].info
    assert info()["text"] == "♫ Rick Astley - Never Gonna Give You Up"

    # Trigger next item and wait for update poll
    topbar.fake_button_press(0, "top", 0, 0, button=4)
    manager_nospawn.c.widget["moc"].eval("self.update(self.poll())")
    assert info()["text"] == "♫ Neil Diamond - Sweet Caroline"

    # Trigger next item and wait for update poll
    # This item's state is set to "STOP" so there's no track title
    topbar.fake_button_press(0, "top", 0, 0, button=4)
    manager_nospawn.c.widget["moc"].eval("self.update(self.poll())")
    assert info()["text"] == "♫"

    # Click to play it and get the information
    topbar.fake_button_press(0, "top", 0, 0, button=1)
    manager_nospawn.c.widget["moc"].eval("self.update(self.poll())")
    assert info()["text"] == "♫ Tom Jones - It's Not Unusual"

    # Trigger previous item and wait for update poll
    topbar.fake_button_press(0, "top", 0, 0, button=5)
    manager_nospawn.c.widget["moc"].eval("self.update(self.poll())")
    assert info()["text"] == "♫ Neil Diamond - Sweet Caroline"


def test_moc_error_handling(patched_moc):
    MockMocpProcess.is_error = True
    # Widget does nothing with error message so text is blank
    assert patched_moc.poll() == ""



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import sys
from types import ModuleType

import pytest

import libqtile.config
from libqtile import widget


class MockMPD(ModuleType):
    class ConnectionError(Exception):
        pass

    class CommandError(Exception):
        pass

    class MPDClient:
        tracks = [
            {"title": "Never gonna give you up", "artist": "Rick Astley", "song": "0"},
            {"title": "Sweet Caroline", "artist": "Neil Diamond"},
            {"title": "Marea", "artist": "Fred Again.."},
            {},
            {"title": "Sweden", "performer": "C418"},
        ]

        def __init__(self):
            self._index = 0
            self._connected = False
            self._state_override = True
            self._status = {"state": "pause"}

        @property
        def _current_song(self):
            return self.tracks[self._index]

        def ping(self):
            if not self._connected:
                raise ConnectionError()
            return self._state_override

        def connect(self, host, port):
            return True

        def command_list_ok_begin(self):
            pass

        def status(self):
            return self._status

        def currentsong(self):
            return self._index + 1

        def command_list_end(self):
            return (self.status(), self._current_song)

        def close(self):
            pass

        def disconnect(self):
            pass

        def pause(self):
            self._status["state"] = "pause"

        def play(self):
            self._status["state"] = "play"

        def stop(self):
            self._status["state"] = "stop"

        def next(self):
            self._index = (self._index + 1) % len(self.tracks)

        def previous(self):
            self._index = (self._index - 1) % len(self.tracks)

        def add_states(self):
            self._status.update(
                {"repeat": "1", "random": "1", "single": "1", "consume": "1", "updating_db": "1"}
            )

        def force_idle(self):
            self._status["state"] = "stop"
            self._index = 3


@pytest.fixture
def mpd2_manager(manager_nospawn, monkeypatch, minimal_conf_noscreen, request):
    # monkeypatch.setattr("libqtile.widget.mpd2widget.MPDClient", MockMPDClient)
    monkeypatch.setitem(sys.modules, "mpd", MockMPD("mpd"))

    config = minimal_conf_noscreen
    config.screens = [
        libqtile.config.Screen(
            top=libqtile.bar.Bar(
                [widget.Mpd2(**getattr(request, "param", dict()))],
                50,
            ),
        )
    ]

    manager_nospawn.start(config)
    yield manager_nospawn


def test_mpd2_widget_display_and_actions(mpd2_manager):
    widget = mpd2_manager.c.widget["mpd2"]
    assert widget.info()["text"] == "⏸ Rick Astley/Never gonna give you up [-----]"

    # Button 1 toggles state
    mpd2_manager.c.bar["top"].fake_button_press(0, "top", 0, 0, 1)
    widget.eval("self.update(self.poll())")
    assert widget.info()["text"] == "▶ Rick Astley/Never gonna give you up [-----]"

    # Button 3 stops
    mpd2_manager.c.bar["top"].fake_button_press(0, "top", 0, 0, 3)
    widget.eval("self.update(self.poll())")
    assert widget.info()["text"] == "■ Rick Astley/Never gonna give you up [-----]"

    # Button 1 toggles state
    mpd2_manager.c.bar["top"].fake_button_press(0, "top", 0, 0, 1)
    widget.eval("self.update(self.poll())")
    assert widget.info()["text"] == "▶ Rick Astley/Never gonna give you up [-----]"

    mpd2_manager.c.bar["top"].fake_button_press(0, "top", 0, 0, 1)
    widget.eval("self.update(self.poll())")
    assert widget.info()["text"] == "⏸ Rick Astley/Never gonna give you up [-----]"

    # Button 5 is "next"
    mpd2_manager.c.bar["top"].fake_button_press(0, "top", 0, 0, 5)
    widget.eval("self.update(self.poll())")
    assert widget.info()["text"] == "⏸ Neil Diamond/Sweet Caroline [-----]"

    mpd2_manager.c.bar["top"].fake_button_press(0, "top", 0, 0, 5)
    widget.eval("self.update(self.poll())")
    assert widget.info()["text"] == "⏸ Fred Again../Marea [-----]"

    # Button 4 is previous
    mpd2_manager.c.bar["top"].fake_button_press(0, "top", 0, 0, 4)
    widget.eval("self.update(self.poll())")
    assert widget.info()["text"] == "⏸ Neil Diamond/Sweet Caroline [-----]"

    mpd2_manager.c.bar["top"].fake_button_press(0, "top", 0, 0, 4)
    widget.eval("self.update(self.poll())")
    assert widget.info()["text"] == "⏸ Rick Astley/Never gonna give you up [-----]"


def test_mpd2_widget_extra_info(mpd2_manager):
    """Quick test to check extra info is displayed ok."""
    widget = mpd2_manager.c.widget["mpd2"]

    # Inject everything to make test quicker
    widget.eval("self.client.add_states()")

    # Update widget and check text
    widget.eval("self.update(self.poll())")
    assert widget.info()["text"] == "⏸ Rick Astley/Never gonna give you up [rz1cU]"


def test_mpd2_widget_idle_message(mpd2_manager):
    """Quick test to check idle message."""
    widget = mpd2_manager.c.widget["mpd2"]

    # Inject everything to make test quicker
    widget.eval("self.client.force_idle()")

    # Update widget and check text
    widget.eval("self.update(self.poll())")
    assert widget.info()["text"] == "■ MPD IDLE[-----]"


@pytest.mark.parametrize(
    "mpd2_manager", [{"status_format": "{currentsong}: {artist}/{title}"}], indirect=True
)
def test_mpd2_widget_current_song(mpd2_manager):
    """Quick test to check currentsong info"""
    widget = mpd2_manager.c.widget["mpd2"]
    assert widget.info()["text"] == "1: Rick Astley/Never gonna give you up"


@pytest.mark.parametrize(
    "mpd2_manager",
    [{"undefined_value": "Unknown", "status_format": "{title} ({year})"}],
    indirect=True,
)
def test_mpd2_widget_custom_undefined_value(mpd2_manager):
    """Quick test to check undefined_value option"""
    widget = mpd2_manager.c.widget["mpd2"]
    assert widget.info()["text"] == "Never gonna give you up (Unknown)"


def test_mpd2_widget_dynamic_artist_value(mpd2_manager):
    """Quick test to check dynamic artist value"""
    widget = mpd2_manager.c.widget["mpd2"]
    widget.eval("self.client._index = 4")
    widget.eval("self.update(self.poll())")
    assert widget.info()["text"] == "⏸ C418/Sweden [-----]"



# Copyright (c) 2021 Matt Colligan
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import typing

from pywayland.server import Listener
from wlroots.util.box import Box
from wlroots.util.clock import Timespec
from wlroots.util.edges import Edges
from wlroots.wlr_types.xdg_shell import XdgSurface, XdgToplevelWMCapabilities

from libqtile import hook
from libqtile.backend import base
from libqtile.backend.base import FloatStates
from libqtile.backend.wayland.window import Static, Window
from libqtile.command.base import expose_command
from libqtile.log_utils import logger

try:
    # Continue if ffi not built, so that docs can be built without wayland deps.
    from libqtile.backend.wayland._ffi import ffi, lib
except ModuleNotFoundError:
    pass

if typing.TYPE_CHECKING:
    from typing import Any

    from libqtile.backend.wayland.core import Core
    from libqtile.core.manager import Qtile
    from libqtile.utils import ColorsType

EDGES_TILED = Edges.TOP | Edges.BOTTOM | Edges.LEFT | Edges.RIGHT
WM_CAPABILITIES = (
    XdgToplevelWMCapabilities.MAXIMIZE
    | XdgToplevelWMCapabilities.FULLSCREEN
    | XdgToplevelWMCapabilities.MINIMIZE
)


class XdgWindow(Window[XdgSurface]):
    """An Wayland client connecting via the xdg shell."""

    def __init__(self, core: Core, qtile: Qtile, surface: XdgSurface):
        Window.__init__(self, core, qtile, surface)

        self._wm_class = surface.toplevel.app_id
        self._geom = Box(0, 0, 0, 0)
        surface.set_wm_capabilities(WM_CAPABILITIES)
        surface.data = self.data_handle
        self.tree = core.scene.xdg_surface_create(self.container, surface)

        self.add_listener(surface.surface.map_event, self._on_map)
        self.add_listener(surface.surface.unmap_event, self._on_unmap)
        self.add_listener(surface.surface.commit_event, self._on_commit)
        self.add_listener(surface.destroy_event, self._on_destroy)
        self.add_listener(surface.toplevel.request_maximize_event, self._on_request_maximize)
        self.add_listener(surface.toplevel.request_fullscreen_event, self._on_request_fullscreen)

        self.ftm_handle = core.foreign_toplevel_manager_v1.create_handle()

    def _on_commit(self, _listener: Listener, _data: Any) -> None:
        if self not in self.core.pending_windows and self.container.node.enabled:
            self.place(
                self.x, self.y, self.width, self.height, self.borderwidth, self.bordercolor
            )

    def _on_request_fullscreen(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: xdgwindow request_fullscreen")
        if self.qtile.config.auto_fullscreen:
            requested = self.surface.toplevel.requested.fullscreen
            if self.fullscreen == requested:
                self.surface.schedule_configure()
            else:
                self.fullscreen = requested
        else:
            # Per xdg-shell protocol we must send a configure in response to this
            # request. Since we're ignoring it, we must schedule a configure manually.
            self.surface.schedule_configure()

    def _on_request_maximize(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: xdgwindow request_maximize")
        self.maximized = self.surface.toplevel.requested.maximized

    def _on_set_title(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: xdgwindow set_title")
        title = self.surface.toplevel.title
        if title and title != self.name:
            self.name = title
            if self.ftm_handle:
                self.ftm_handle.set_title(self.name)
            hook.fire("client_name_updated", self)

    def _on_set_app_id(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: xdgwindow set_app_id")
        self._wm_class = self.surface.toplevel.app_id
        if self.ftm_handle:
            self.ftm_handle.set_app_id(self._wm_class or "")

    def unhide(self) -> None:
        self._wm_class = self.surface.toplevel.app_id

        if self not in self.core.pending_windows:
            # Regular usage
            if not self.container.node.enabled and self.group and self.group.screen:
                self.container.node.set_enabled(enabled=True)
            return

        # This is the first time this window has mapped, so we need to do some initial
        # setup.
        self.core.pending_windows.remove(self)
        self._wid = self.core.new_wid()
        logger.debug(
            "Managing new top-level window with window ID: %s, app_id: %s",
            self._wid,
            self._wm_class,
        )

        # Save the client's desired geometry
        surface = self.surface
        geometry = surface.get_geometry()
        self._width = self._float_width = geometry.width
        self._height = self._float_height = geometry.height

        # Tell the client to render tiled edges
        surface.set_tiled(EDGES_TILED)

        handle = self.ftm_handle
        assert handle is not None

        # Get the client's name
        if surface.toplevel.title:
            self.name = surface.toplevel.title
            handle.set_title(self.name)
        if self._wm_class:
            handle.set_app_id(self._wm_class or "")

        # Add the toplevel's listeners
        self.add_listener(surface.toplevel.set_title_event, self._on_set_title)
        self.add_listener(surface.toplevel.set_app_id_event, self._on_set_app_id)
        self.add_listener(handle.request_maximize_event, self._on_foreign_request_maximize)
        self.add_listener(handle.request_minimize_event, self._on_foreign_request_minimize)
        self.add_listener(handle.request_activate_event, self._on_foreign_request_activate)
        self.add_listener(handle.request_fullscreen_event, self._on_foreign_request_fullscreen)
        self.add_listener(handle.request_close_event, self._on_foreign_request_close)

        self.qtile.manage(self)

        # Send a frame done event to provide an opportunity to redraw even if we aren't
        # going to map (i.e. because the window was opened on a hidden group)
        surface.surface.send_frame_done(Timespec.get_monotonic_time())

    @expose_command()
    def kill(self) -> None:
        self.surface.send_close()

    def has_fixed_size(self) -> bool:
        state = self.surface.toplevel._ptr.current
        return 0 < state.min_width == state.max_width and 0 < state.min_height == state.max_height

    def is_transient_for(self) -> base.WindowType | None:
        """What window is this window a transient window for?"""
        if parent := self.surface.toplevel.parent:
            for win in self.qtile.windows_map.values():
                if isinstance(win, XdgWindow) and win.surface.toplevel == parent:
                    return win
        return None

    def get_pid(self) -> int:
        pid = ffi.new("pid_t *")
        lib.wl_client_get_credentials(self.surface._ptr.client.client, pid, ffi.NULL, ffi.NULL)
        return pid[0]

    def _update_fullscreen(self, do_full: bool) -> None:
        if do_full != (self._float_state == FloatStates.FULLSCREEN):
            self.surface.set_fullscreen(do_full)
            if self.ftm_handle:
                self.ftm_handle.set_fullscreen(do_full)

    def handle_activation_request(self, focus_on_window_activation: str) -> None:
        """Respond to XDG activation requests targeting this window."""
        assert self.qtile is not None

        if self.group is None:
            # Likely still pending, ignore this request.
            return

        if focus_on_window_activation == "focus":
            logger.debug("Focusing window (focus_on_window_activation='focus')")
            self.qtile.current_screen.set_group(self.group)
            self.group.focus(self)

        elif focus_on_window_activation == "smart":
            if not self.group.screen:
                logger.debug("Ignoring focus request (focus_on_window_activation='smart')")
            elif self.group.screen == self.qtile.current_screen:
                logger.debug("Focusing window (focus_on_window_activation='smart')")
                self.qtile.current_screen.set_group(self.group)
                self.group.focus(self)
            else:
                self._urgent = True
                hook.fire("client_urgent_hint_changed", self)

        elif focus_on_window_activation == "urgent":
            self._urgent = True
            hook.fire("client_urgent_hint_changed", self)

    def clip(self) -> None:
        if not self.tree:
            return
        if not self.tree.node.enabled:
            return
        if next(self.tree.children, None) is None:
            return
        self.tree.node.subsurface_tree_set_clip(
            Box(self._geom.x, self._geom.y, self.width, self.height)
        )

    def place(
        self,
        x: int,
        y: int,
        width: int,
        height: int,
        borderwidth: int,
        bordercolor: ColorsType | None,
        above: bool = False,
        margin: int | list[int] | None = None,
        respect_hints: bool = False,
    ) -> None:
        # Adjust the placement to account for layout margins, if there are any.
        if margin is not None:
            if isinstance(margin, int):
                margin = [margin] * 4
            x += margin[3]
            y += margin[0]
            width -= margin[1] + margin[3]
            height -= margin[0] + margin[2]

        if respect_hints:
            state = self.surface.toplevel._ptr.current
            width = max(width, state.min_width)
            height = max(height, state.min_height)
            if state.max_width:
                width = min(width, state.max_width)
            if state.max_height:
                height = min(height, state.max_height)

        # save x and y float offset
        if self.group is not None and self.group.screen is not None:
            self.float_x = x - self.group.screen.x
            self.float_y = y - self.group.screen.y

        if width < 1:
            width = 1

        if height < 1:
            height = 1

        geom = self.surface.get_geometry()
        place_changed = any(
            [self.x != x, self.y != y, self._width != width, self._height != height]
        )
        geom_changed = any(
            [
                self._geom.x != geom.x,
                self._geom.y != geom.y,
                self._geom.width != geom.width,
                self._geom.height != geom.height,
            ]
        )
        needs_repos = place_changed or geom_changed
        has_border_changed = any(
            [borderwidth != self.borderwidth, bordercolor != self.bordercolor]
        )

        self._geom = geom
        self.x = x
        self.y = y
        self._width = width
        self._height = height

        self.container.node.set_position(x, y)
        if needs_repos:
            self.surface.set_size(width, height)
            self.surface.set_bounds(width, height)
            self.clip()

        if needs_repos or has_border_changed:
            self.paint_borders(bordercolor, borderwidth)

        if above:
            self.bring_to_front()

    @expose_command()
    def static(
        self,
        screen: int | None = None,
        x: int | None = None,
        y: int | None = None,
        width: int | None = None,
        height: int | None = None,
    ) -> None:
        Window.static(self, screen, x, y, width, height)
        win = typing.cast(XdgStatic, self.qtile.windows_map[self._wid])

        for pc in self.core.pointer_constraints.copy():
            if pc.window is self:
                pc.window = win

        hook.fire("client_managed", win)

    def _to_static(
        self, x: int | None, y: int | None, width: int | None, height: int | None
    ) -> XdgStatic:
        return XdgStatic(
            self.core,
            self.qtile,
            self,
        )


class XdgStatic(Static[XdgSurface]):
    """A static window belonging to the XDG shell."""

    def __init__(
        self,
        core: Core,
        qtile: Qtile,
        win: XdgWindow,
    ):
        surface = win.surface
        Static.__init__(self, core, qtile, surface, win.wid)

        if surface.toplevel.title:
            self.name = surface.toplevel.title
        self._wm_class = surface.toplevel.app_id

        self.add_listener(surface.surface.map_event, self._on_map)
        self.add_listener(surface.surface.unmap_event, self._on_unmap)
        self.add_listener(surface.destroy_event, self._on_destroy)
        # self.add_listener(surface.surface.commit_event, self._on_commit)
        self.add_listener(surface.toplevel.set_title_event, self._on_set_title)
        self.add_listener(surface.toplevel.set_app_id_event, self._on_set_app_id)

        # Take control of the scene tree
        self.container = win.container
        self.container.node.data = self.data_handle
        self.tree = win.tree

    @expose_command()
    def kill(self) -> None:
        self.surface.send_close()

    def hide(self) -> None:
        super().hide()
        self.container.node.set_enabled(enabled=False)

    def unhide(self) -> None:
        self.container.node.set_enabled(enabled=True)

    def place(
        self,
        x: int,
        y: int,
        width: int,
        height: int,
        borderwidth: int,
        bordercolor: ColorsType | None,
        above: bool = False,
        margin: int | list[int] | None = None,
        respect_hints: bool = False,
    ) -> None:
        self.x = x
        self.y = y
        self._width = width
        self._height = height
        self.surface.set_size(width, height)
        self.surface.set_bounds(width, height)
        self.container.node.set_position(x, y)

    def _on_set_title(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: xdgstatic set_title")
        title = self.surface.toplevel.title
        if title and title != self.name:
            self.name = title
            if self.ftm_handle:
                self.ftm_handle.set_title(self.name)
            hook.fire("client_name_updated", self)

    def _on_set_app_id(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: xdgstatic set_app_id")
        self._wm_class = self.surface.toplevel.app_id
        if self.ftm_handle:
            self.ftm_handle.set_app_id(self._wm_class or "")
